# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json

import keystoneclient.v2_0.client
from oslo.config import cfg
import six

opts = [
    cfg.StrOpt('user',
               default='accounting',
               help='User to authenticate as.'),
    cfg.StrOpt('password',
               default='',
               help='Password to authenticate with.'),
    cfg.StrOpt('endpoint',
               default='',
               help='Keystone endpoint to autenticate with.'),
    cfg.BoolOpt('insecure',
                default=False,
                help='Perform an insecure connection (i.e. do '
                'not verify the server\'s certificate. DO NOT USE '
                'IN PRODUCTION.'),
    cfg.StrOpt('mapping_file',
               default='/etc/caso/voms.json',
               help='File containing the VO <-> tenant mapping as used '
               'in Keystone-VOMS.'),
]

CONF = cfg.CONF
CONF.register_opts(opts, group="extractor")

openstack_vm_statuses = {
    'active': 'started',
    'build': 'started',
    'confirming_resize': 'started',
    'deleted': 'completed',
    'error': 'error',
    'hard_reboot': 'started',
    'migrating': 'started',
    'password': 'started',
    'paused': 'paused',
    'reboot': 'started',
    'rebuild': 'started',
    'rescue': 'started',
    'resize': 'started',
    'revert_resize': 'started',
    'verify_resize': 'started',
    'shutoff': 'completed',
    'suspended': 'suspended',
    'terminated': 'completed',
    'stopped': 'stopped',
    'saving': 'started',
    'unknown': 'unknown',
}


@six.add_metaclass(abc.ABCMeta)
class BaseExtractor(object):
    def __init__(self):
        try:
            mapping = json.loads(open(CONF.extractor.mapping_file).read())
        except ValueError:
            # FIXME(aloga): raise a proper exception here
            raise
        else:
            self.voms_map = {}
            for vo, vomap in mapping.iteritems():
                self.voms_map[vomap["tenant"]] = vo

    def _get_keystone_client(self, tenant):
        client = keystoneclient.v2_0.client.Client(
            username=CONF.extractor.user,
            password=CONF.extractor.password,
            tenant_name=tenant,
            auth_url=CONF.extractor.endpoint,
            insecure=CONF.extractor.insecure)
        client.authenticate()
        return client

    def _get_keystone_users(self, ks_client):
        tenant_id = ks_client.tenant_id
        users = ks_client.users.list(tenant_id=tenant_id)
        return {u.id: u.name for u in users}

    def vm_status(self, status):
        return openstack_vm_statuses.get(status.lower(), 'unknown')



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import dateutil.parser
from dateutil import tz
import novaclient.client
from oslo.config import cfg

from caso.extract import base
from caso import record

CONF = cfg.CONF
CONF.import_opt("site_name", "caso.extract.manager")
CONF.import_opt("user", "caso.extract.base", "extractor")
CONF.import_opt("password", "caso.extract.base", "extractor")
CONF.import_opt("endpoint", "caso.extract.base", "extractor")
CONF.import_opt("insecure", "caso.extract.base", "extractor")


class OpenStackExtractor(base.BaseExtractor):
    def _get_conn(self, tenant):
        client = novaclient.client.get_client_class("2")
        conn = client(
            CONF.extractor.user,
            CONF.extractor.password,
            tenant,
            CONF.extractor.endpoint,
            insecure=CONF.extractor.insecure,
            service_type="compute")
        conn.authenticate()
        return conn

    def extract_for_tenant(self, tenant, lastrun):
        # Some API calls do not expect a TZ, so we have to remove the timezone
        # from the dates. We assume that all dates coming from upstream are
        # in UTC TZ.
        lastrun = lastrun.replace(tzinfo=None)
        now = datetime.datetime.now(tz.tzutc()).replace(tzinfo=None)
        end = now + datetime.timedelta(days=1)

        # Try and except here
        conn = self._get_conn(tenant)
        ks_conn = self._get_keystone_client(tenant)
        users = self._get_keystone_users(ks_conn)
        tenant_id = conn.client.tenant_id
        servers = conn.servers.list(search_opts={"changes-since": lastrun})

        # FIXME(aloga): use start and end from the retreived servers
        aux = conn.usage.get(tenant_id, lastrun, end)
        usages = getattr(aux, "server_usages", [])

        images = conn.images.list()
        records = {}

        vo = self.voms_map.get(tenant)

        for server in servers:
            status = self.vm_status(server.status)
            image_id = None
            for image in images:
                if image.id == server.image['id']:
                    image_id = image.metadata.get("vmcatcher_event_ad_mpuri",
                                                  None)
                    break

            if image_id is None:
                image_id = server.image['id']

            r = record.CloudRecord(server.id,
                                   CONF.site_name,
                                   server.name,
                                   server.user_id,
                                   server.tenant_id,
                                   vo,
                                   cloud_type="OpenStack",
                                   status=status,
                                   image_id=image_id,
                                   user_dn=users.get(server.user_id, None))
            records[server.id] = r

        for usage in usages:
            if usage["instance_id"] not in records:
                continue
            instance_id = usage["instance_id"]
            records[instance_id].memory = usage["memory_mb"]
            records[instance_id].cpu_count = usage["vcpus"]
            records[instance_id].disk = usage["local_gb"]
            records[instance_id].cpu_duration = int(usage["hours"] * 3600)

            started = dateutil.parser.parse(usage["started_at"])
            records[instance_id].start_time = int(started.strftime("%s"))
            if usage.get("ended_at", None) is not None:
                ended = dateutil.parser.parse(usage['ended_at'])
                records[instance_id].end_time = int(ended.strftime("%s"))
                wall = ended - started
            else:
                wall = now - started

            records[instance_id].wall_duration = int(wall.total_seconds())
        return records



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import dateutil.parser
from dateutil import tz
from oslo.config import cfg
from oslo.utils import importutils
import six

from caso import log

SUPPORTED_EXTRACTORS = {
    'nova': 'caso.extract.nova.OpenStackExtractor',
    'ceilometer': 'caso.extract.ceilometer.CeilometerExtractor',
}

opts = [
    cfg.StrOpt('site_name',
               help='Site name as in GOCDB.'),
    cfg.ListOpt('tenants',
                default=[],
                help='List of tenants to extract accounting records from.'),
]

cli_opts = [
    cfg.StrOpt('extract_from',
               help='Extract records from this date. If it is not set, '
               'extract records from last run. If none are set, extract '
               'records from the beginning of time. If no time zone is '
               'specified, UTC will be used.'),
    cfg.StrOpt('extractor',
               choices=SUPPORTED_EXTRACTORS,
               default='nova',
               help=('Which extractor to use for getting the data. '
                     'Only the following middlewares are supported: %s. '
                     'If you do not specify anything, nova will be '
                     'used.' % SUPPORTED_EXTRACTORS.keys())),
]

CONF = cfg.CONF

CONF.register_opts(opts)
CONF.register_cli_opts(cli_opts)

LOG = log.getLogger(__name__)


class Manager(object):
    def __init__(self):
        extractor_class = importutils.import_class(
            SUPPORTED_EXTRACTORS[CONF.extractor])
        self.extractor = extractor_class()
        self.records = None

    def _extract(self, extract_from):
        self.records = {}
        for tenant in CONF.tenants:
            records = self.extractor.extract_for_tenant(tenant,
                                                        extract_from)
            LOG.info("Extracted %d records for tenant '%s' from %s to now" %
                     (len(records), tenant, extract_from))
            self.records.update(records)

    def get_records(self, lastrun="1970-01-01"):
        extract_from = CONF.extract_from or lastrun

        if isinstance(extract_from, six.string_types):
            extract_from = dateutil.parser.parse(extract_from)

        if extract_from.tzinfo is None:
            extract_from.replace(tzinfo=tz.tzutc())
        if self.records is None:
            self._extract(extract_from)
        return self.records



# -*- coding: utf-8 -*-

# Copyright 2014 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import ceilometerclient.client
import dateutil.parser
from dateutil import tz
import glanceclient.client
from oslo.config import cfg

from caso.extract import base
from caso import log
from caso import record

CONF = cfg.CONF
CONF.import_opt("site_name", "caso.extract.manager")
CONF.import_opt("user", "caso.extract.base", "extractor")
CONF.import_opt("password", "caso.extract.base", "extractor")
CONF.import_opt("endpoint", "caso.extract.base", "extractor")
CONF.import_opt("insecure", "caso.extract.base", "extractor")

LOG = log.getLogger(__name__)


class CeilometerExtractor(base.BaseExtractor):
    def _get_ceilometer_client(self, tenant):
        return ceilometerclient.client.get_client(
            '2',
            os_auth_url=CONF.extractor.endpoint,
            os_username=CONF.extractor.user,
            os_password=CONF.extractor.password,
            os_tenant_name=tenant,
            insecure=CONF.extractor.insecure)

    def _get_glance_client(self, ks_client):
        glance_ep = ks_client.service_catalog.get_endpoints(
            service_type='image',
            endpoint_type='public')
        glance_url = glance_ep['image'][0]['publicURL']
        # glance client does not seem to work with user/password
        return glanceclient.client.Client(
            '2',
            glance_url,
            token=ks_client.auth_token,
            insecure=CONF.extractor.insecure)

    def _build_query(self, project_id=None, start_date=None, end_date=None):
        q = []
        if project_id:
            q.append({'field': 'project_id', 'value': project_id})
        if start_date:
            q.append({'field': 'timestamp', 'op': 'ge', 'value': start_date})
        if end_date:
            q.append({'field': 'timestamp', 'op': 'le', 'value': end_date})
        return q

    def _fill_metric(self, metric_name, samples, records,
                     get_id=lambda s: s.resource_id,
                     unit_conv=lambda v: v):
        """Fills a given metric in the records

        get_id is a function that gets the instance id from the sample
        conv is a function that converts the metric to the units
             requested by the usage record

        """

        instance_timestamps = {}
        the_past = datetime.datetime(1, 1, 1)
        for sample in samples:
            instance_id = get_id(sample)
            r = records.get(instance_id)
            if not r:
                # XXX: there is a sample for a VM that has no instance sample?
                LOG.debug("VM with some usage info but no instance metric?!")
                continue
            instance_ts = instance_timestamps.get(instance_id, the_past)
            sample_ts = dateutil.parser.parse(sample.timestamp)
            if sample_ts > instance_ts:
                r.__dict__[metric_name] = unit_conv(sample.counter_volume)
                instance_timestamps[instance_id] = sample_ts

    def _fill_cpu_metric(self, cpu_samples, records):
        self._fill_metric('cpu_duration', cpu_samples, records,
                          # convert ns to s
                          unit_conv=lambda v: int(v / 1e9))

    def _fill_net_metric(self, metric_name, net_samples, records):
        self._fill_metric(metric_name, net_samples, records,
                          lambda s: s.resource_metadata.get('instance_id'),
                          # convert bytes to GB
                          unit_conv=lambda v: int(v / 2 ** 30))

    def _get_time_field(self, metadata, fields=[]):
        for f in fields:
            t = metadata.get(f, None)
            if t:
                dateutil.parser.parse(t).replace(tzinfo=None)
        else:
            return None

    def _build_record(self, instance, users, vo, images, now):
        metadata = instance.resource_metadata
        status = self.vm_status(metadata.get('state', ''))
        instance_image_id = metadata.get('image.id')
        if not instance_image_id:
            instance_image_id = metadata.get('image_meta.base_image_ref')
        image_id = None
        for image in images:
            if instance_image_id == image.id:
                image_id = image.get('vmcatcher_event_ad_mpuri', None)
                break
        else:
            image_id = instance_image_id
        r = record.CloudRecord(instance.resource_id,
                               CONF.site_name,
                               metadata.get('display_name'),
                               instance.user_id,
                               instance.project_id,
                               vo,
                               cpu_count=metadata.get('vcpus'),
                               memory=metadata.get('memory_gb'),
                               disk=metadata.get('disk_gb'),
                               # this should contain "caso ceilometer vX"
                               cloud_type="OpenStack",
                               status=status,
                               image_id=image_id,
                               user_dn=users.get(instance.user_id, None))

        start_time = self._get_time_field(metadata,
                                          ('launched_at', 'created_at'))
        end_time = self._get_time_field(metadata,
                                        ('terminated_at', 'deleted_at'))
        if start_time:
            r.start_time = int(start_time.strftime("%s"))
            if end_time:
                r.end_time = int(end_time.strftime("%s"))
                wall = end_time - start_time
                r.wall_duration = int(wall.total_seconds())
            else:
                wall = now - start_time
                r.wall_duration = int(wall.total_seconds())
        return r

    def extract_for_tenant(self, tenant, lastrun):
        now = datetime.datetime.now(tz.tzutc())
        now.replace(tzinfo=None)

        # Try and except here
        # Getting clients
        ks_client = self._get_keystone_client(tenant)
        client = self._get_ceilometer_client(tenant)
        glance_client = self._get_glance_client(ks_client)

        # users
        users = self._get_keystone_users(ks_client)
        tenant_id = ks_client.tenant_id

        search_query = self._build_query(tenant_id, lastrun)
        instances = client.samples.list('instance', search_query)

        # XXX should we query glance for every VM or not?
        images = list(glance_client.images.list())

        records = {}

        vo = self.voms_map.get(tenant)

        for instance in instances:
            # it seems the only event type with relevant metadata is
            # 'compute.instance.exists'
            ev_type = instance.resource_metadata.get('event_type', None)
            if ev_type != 'compute.instance.exists':
                continue
            # this assumes that records are returned with decreasing timestamp
            # so the status and dates are the last ones.
            if instance.resource_id in records:
                continue
            records[instance.resource_id] = self._build_record(instance,
                                                               users,
                                                               vo,
                                                               images,
                                                               now)

        cpu = client.samples.list(meter_name='cpu', q=search_query)
        self._fill_cpu_metric(cpu, records)
        net_in = client.samples.list(meter_name='network.incoming.bytes',
                                     q=search_query)
        self._fill_net_metric('network_in', net_in, records)
        net_out = client.samples.list(meter_name='network.outcoming.bytes',
                                      q=search_query)
        self._fill_net_metric('network_out', net_out, records)

        return records
